# $Id: Reader.pm 1941 2008-07-08 10:38:42Z clerger $
package Forest::Dependency::Reader;

require 5.005;
use strict;
use warnings;
use Forest;
use XML::Parser;
use UNIVERSAL qw( isa ) ;

our @ISA = qw{XML::Parser};

our $forest;
our %tmp;
our @entries;


sub new {
  my $proto = shift;
  my $class = ref($proto) || $proto;
  my $self = $class->SUPER::new(Pkg=>__PACKAGE__,
				Style=>'Subs',
				@_);
  $self->setHandlers(Final => \&Final,
		     Init => \&Init,
		     Char => \&Char
		    );
  return $self;
}


sub parse {
  my $self = __PACKAGE__->new;
  return $self->SUPER::parse(@_);
}

sub Char {
  my $self = shift;
  my $text = shift;
  $text =~ s/^\s+//o;
  $text =~ s/\s+$//o;
  $self->{Current}{$self->{'Where'}}=$text unless ($text  =~ /^\s*$/);
}

sub Final {
  my $self = shift;
  delete $self->{Current};
  delete $self->{Where};
  delete $self->{Stack};
  foreach my $oid (keys %{$self->{Forest}{op}}) {
    my $op = $self->{Forest}{op}{$oid};
    unless (exists $op->{type}) {
      ## Desambiguation on robust parses may return
      ## nodes not belonging to any derivation
      ## => pseudo derivations are therefore introduced
      ## maybe the problem should better be solved during the desambiguation (easyforest)
      printf STDERR "Filling incomplete op $oid lex=$op->{lex} span=@{$op->{span}}\n";
      $op->{type} = 'lexical';
      $op->{tree} = 'lexical';
      my $id = ++$self->{Forest}{nb};
      $id = "pseudo$id";
      $self->{Forest}{op}{$id} = 
	Forest::Op->new( id => $id,
			 cat => 'pseudo',
			 lex => '',
			 token => '',
			 lemma => '',
			 type => 'subst',
			 span => $op->{span},
			 derivations => [ Forest::Deriv->new( tree => 'pseudo',
							      nodes => { void => [$oid] }
							    )
					]
		       );
    }
  }
  return $self->{Forest};
}

sub Init {
  my $self = shift;
  $self->{Forest} = Forest->new( {'op'=> {}, 'nb' => 0} );
  $self->{cluster} = {};
  $self->{deriv} = {};
  $self->{op} = {};
}


sub dependencies {
  my $self = shift;
}

sub _dependencies {
}

=head1 METHODS

=head2 cluster

A cluster provides information about a portion of the input sentence 
precisely between positions left and right

=cut

sub cluster {
  my $self = shift;
  my $tag = shift;
  my %attr = @_;
  $self->{cluster}{$attr{id}}
    = { lex => $attr{lex},
	token => $attr{token} || $attr{lex},
	span =>  [$attr{left},$attr{right}] };
}

=head2 node

Node elements are used to denote anchors and other lexicals
nodes are then covered by maximal non terminals 'op'
A node refers a cluster

=cut

sub node {
  my $self = shift;
  my $tag = shift;
  my %attr = @_;
  my $id = $attr{id};
  my $cluster = $self->{cluster}{$attr{cluster}};
  my $lex = $cluster->{lex};
  my $oid;
  my $tree = $attr{tree};
  unless ($lex eq '') {
    ## the node is not a virtual one ($lex eq '')
    ## but is not necessarily an anchor node
    ## it may also be a lexical or epsilon
    $oid = $self->{op}{$id} ||= ++$self->{Forest}{nb};
    my $op = $self->{Forest}{op}{$oid} 
      = Forest::Op->new( id => $oid,
			 derivations => [],
			 cat => $attr{cat},
			 lex => $lex,
			 token => $attr{form} || $cluster->{token},
			 span => $cluster->{span}
		       );
    $op->{lemma} = $attr{lemma} if (exists $attr{lemma});
##    print STDERR "Register node $oid: span=@{$op->{span}} les=$op->{lex} cat=$op->{cat}\n"
  }
  if (exists $attr{deriv}) {
    if (defined $oid) {		
      # A node represents an anchor is has a deriv, a tree and a lex
      $self->{Forest}{op}{$oid}{type} = 'anchor';
      $self->{Forest}{op}{$oid}{tree} = $tree;
##      print STDERR "Register anchor $oid\n";
    }
    foreach my $did (split(/\s+/,$attr{deriv})) {
      ## each deriv is associated to a Forest::Deriv whose
      ## tree is $tree and anchor is $op
      ## The derive included in some Forest::Op whose cat is $xcat
      $did =~ s/_\d+$//;
      next if (exists $self->{deriv}{$did});
      my $deriv = $self->{deriv}{$did}
	= Forest::Deriv->new( tree => $tree,
			      nodes => {}
			    );
      $deriv->{nodes}{$attr{cat}} = [$oid] unless ($lex eq '');
    }
  } 
}

=head2 edge

Edges relate two nodes and provide additionnal info

=cut

sub edge {
  my $self = shift;
  my $tag = shift;
  my %attr = @_;

  my $sid = $self->{op}{$attr{source}};
  my $tid = $self->{op}{$attr{target}};
  
  my $edge = { source => $sid,
	       target => $tid,
	       label => $attr{label},
	       type => $attr{type},
	       sid => $attr{source},
	       tid => $attr{target},
	     };
  $self->push($edge);
}

sub _edge {
  shift->pop();
}

=head2 deriv

derivs are embedded within edges and indicate the op
covering the target node
plus a deriv id

=cut

sub deriv {
  my $self = shift;
  my $tag = shift;
  my %attr = @_;
  my $edge = $self->{Current};
  foreach my $did ($attr{name} || split(/\s+/,$attr{names}) ){
    ## the use of the 'name' attribute is deprecated in favor of 'names'
    ##  my $did = $attr{name};
    $did =~ s/_\d+$//;
    my $deriv = $self->{deriv}{$did};
    my $tid;
    my $tmpid;
    my $type = $edge->{type};
    if ($type eq 'lexical' || $type eq 'epsilon') {
      $tid = $edge->{target};
      $tmpid = $edge->{tid};
      $self->{Forest}{op}{$tid}{type} ||= $type;
    } else {
      $tmpid = $attr{target_op};
      $tid = $self->{op}{$attr{target_op}}
	||= ++$self->{Forest}{nb};
      my $target_op = $self->{Forest}{op}{$tid}
	||= Forest::Op->new( id => $tid,
			     derivations => [],
			   );
      $target_op->{type}=$edge->{type};
    }
    ##  print STDERR "Reader register edge $edge->{sid}:$edge->{source} -[$edge->{label}:$edge->{type}]-> $edge->{tid}:$edge->{target} [max=$tmpid:$tid]\n";
    push(@{$deriv->{nodes}{$edge->{label}}},$tid);
  }
}

=head2 op

op elements are used for maximal non terminals 
information provided by the deriv attribute is not used here
because it is redundant with the info provided by the deriv element
within edges

=cut

sub op {			
  my $self = shift;
  my $tag = shift;
  my %attr = @_;

  my $id = $attr{id};
  my $oid = $self->{op}{$id} ||= ++$self->{Forest}{nb};
  my $op = $self->{Forest}{op}{$oid} 
    ||= Forest::Op->new( id => $oid,
			 derivations => [],
			 type => 'subst'
		       );
  $op->{span} = [ split(/\s+/,$attr{span}) ];
  $op->{cat} = $attr{cat};

  if (exists $attr{deriv}) {
    my %seen=();
    foreach my $did (split(/\s+/,$attr{deriv})) {
      $did =~ s/_\d+$//;
      next if (exists $seen{$did});
      $seen{$did} = 1;
      my $deriv = $self->{deriv}{$did};
      push(@{$op->{derivations}},$deriv);
    }
  }
  $self->push($op);
}

sub _op {
  shift->pop;
}

=head2 hypertag

hypertag elements are used to provide the feature structure
information attached to derivations, and resulting from the
unification of the hypertags of the trees and anchors.

=cut


sub hypertag {			
  my $self = shift;
  my $tag = shift;
  my %attr = @_;

  my @derivs = split(/\s+/,$attr{derivs});
  my $oid = $attr{op};

  ## All derivs under an hypertag share the same base derivation
  my $did = $derivs[0];

  ## we are assuming that everything is OK
  ## ie that the derivs have already been created !
  my $deriv = $self->{deriv}{$did};

  $self->push($deriv);
  $self->{Where} = 'ht';
}

sub _hypertag {
  my $self = shift;
  undef $self->{Where};
  self->pop;
}

######################################################################

sub narg {
    my $self = shift;
    shift;
    my %attr = @_;
    my $node = $self->{Current};
    $self->{Where} = $attr{'type'};
}

sub narg_ {
    my $self = shift;
    undef $self->{Where};
}

## Handling Feature Structures

sub plus {
    my $self = shift;
    my $plus = Forest::Plus->new();
    push( @{ $self->{Current}{'value'} },$plus);
}

sub minus {
    my $self = shift;
    my $minus = Forest::Minus->new();
    push( @{ $self->{Current}{'value'} },$minus);
}

sub fs {
    my $self = shift;
    my $feature = Forest::Feature->new();
    my $where = $self->{Where} || 'feature';
    $feature->{Save_Where} = $self->push_val($feature,$where);
    $self->push($feature);
    $self->{Where} = 'val';
}

sub fs_ {
  my $self = shift;
  $self->{Where} = $self->{Current}{Save_Where};
  delete $self->{Current}{Save_Where};
  $self->pop();
}

sub f {
    my $self = shift;
    my $tag = shift;
    my %attr = @_;
    my $comp = Forest::Comp->new( { comp => $attr{name}, value =>[] } );
    $self->{Current}{'f'}{$comp->{'comp'}} = $comp;
    $self->push($comp);
}

sub f_ {
    shift->pop;
}

sub val {
    my $self = shift;
    my $tag = shift;
    my $val = Forest::Val->new();
    push( @{ $self->{Current}{'value'} },$val);
    $self->push($val);
}

sub val_ {
    shift->pop;
}

sub link {
    my $self = shift;
    my $tag = shift;
    my %attr = @_;
    $self->{Current}{'id'}=$attr{'id'};
}

## utilities
package XML::Parser::Expat;
use UNIVERSAL qw( isa ) ;

sub push {
    my ($self,$obj) = @_;
    CORE::push @{ $self->{Stack} }, $self->{Current};
    $self->{Current} = $obj;
}

sub pop {
    my $self = shift;
    $self->{Current} = CORE::pop @{ $self->{Stack} };
}

sub push_val {
  my $self = shift;
  my $val = shift;
  my $where = $self->{Where} || shift || 'value';
  if (defined $self->{Current}{$where}
      && isa( $self->{Current}{$where},"ARRAY")) {
    CORE::push(@{$self->{Current}{$where}},$val);
  } else {
    $self->{Current}{$where} = $val;
  }
  return $where;
}


1;

__END__

=head1 NAME

Forest::Dependency::Reader - Reading XML DEP files and build an
internal representation of shared forest.

=head1 SYNOPSIS

  use Forest::Dependency::Reader;

=head1 DESCRIPTION

=head2 EXPORT

None by default.

=head1 SEE ALSO

perl(1).

=head1 COPYRIGHT AND LICENSE

Copyright (C) 2000-2007, INRIA.

This program is free software; you can redistribute it and/or modify it under
the same terms as Perl itself.

=head1 AUTHOR

Eric de la Clergerie <Eric.De_La_Clergerie@inria.fr>

=cut

