#$Id: Writer.pm 2065 2008-08-28 11:17:33Z clerger $
package Forest::Dependency::Writer;

require 5.005_62;
use strict;
use warnings;
use Forest;
use Forest::XML::Writer;

our $forest;
our $dependency;
our $pseudoop;

our %color = ( 'subst'   => 'lightskyblue',
	       'adj'     => 'indianred3',
	       'anchor'  => 'yellow',
	       'lexical' => 'plum',
	       'epsilon' => 'green'
	     );

our %trees;

our $dcounter;			# to assign id to derivations
our $htid;
our $nid;
our $eid;

our %derivations;

sub cluster {
    my ($l, $r) = @_;
    return "cluster$l"."_"."$r";
}

sub node {
  my $node  = shift;
  my $label = cluster($node->{pos}[0], $node->{pos}[1])."_$node->{cat}_$node->{lemma}_".norm_tree(treename($node->{tree}));
  $label =~ s/\W//og;
  $label =~ s/_+/_/og;
  return $label;
}

sub treename {
  my $treename = shift;
  return (split(/\s+/,$treename))[0];
}

sub Forest::cluster_loop {
  my $self = shift;
  my $cluster_handler = shift;
   foreach my $posa (sort {$a <=> $b} keys %{$dependency}) {
     foreach my $posb (sort {$a <=> $b} keys %{$dependency->{$posa}}) {
       $cluster_handler->($self, $posa, $posb, $dependency->{$posa}{$posb}, @_);
     }
   }
}

sub Forest::node_loop {
  my ($self, $posa, $posb, $nodes, $node_handler, @l) = @_;
  foreach my $lemmas (values %{$nodes->{tree}}) {
    foreach my $trees (values %$lemmas) {
      foreach my $target (values %$trees) {
	$node_handler->($self, $target,@l);
      }
    }
  }
}

sub Forest::edge_loop {
  my ($self,$target,$edge_handler,@l) = @_;
  while (my ($label,$edges2) = each %{$target->{label}}) {
    foreach my $edge (@$edges2) {
      $edge_handler->($self,$target,$edge,$label,@l);
    }
  }
}

sub Forest::dependency_normalize {
    my $self = shift;
    my $nb   = $self->{'nb'};
    $forest  = $self;
    $dependency = {};
    %trees = ();
    $pseudoop={};

    # First pass
    foreach my $xop (values %{$self->{'op'}}) {
      $xop->dependency();
    }

    # Second pass: emiting
    foreach my $pos (sort {$a <=> $b} keys %{$dependency}) {
	foreach my $posb (sort {$a <=> $b} keys %{$dependency->{$pos}}) {
	    my $lex   = $dependency->{$pos}{$posb}{'lex'};
	    my @trees = keys %{$dependency->{$pos}{$posb}{'tree'}};
	    foreach my $tree (sort @trees) {
		my $cat      = $trees{$tree}{'cat'};
		my $node     = $trees{$tree}{'node'};
		my $edges    = $dependency->{$pos}{$posb}{'tree'}{$tree};
		my $treename = treename($tree);
		foreach my $label (sort keys %{$edges}) {
		    foreach my $target (@{$edges->{$label}}) {
			my $tpos  = $target->{'pos'}[0];
			my $tposb = $target->{'pos'}[1];
			my $ttree = $target->{'tree'};
			my $ttreename = treename($ttree);
			my $tlex  = $dependency->{$tpos}{$tposb}{'lex'};
			my $tcat  = $trees{$ttree}{'cat'};
			my $tnode = $trees{$ttree}{'node'};
			print "$target->{'type'}\t{$pos $posb $lex $treename $node $cat}\t{$tpos $tposb $tlex $ttreename $tnode $tcat}\t$label\n";
		    }
		}
	    }
	}
    }
}

sub tree_emit {
  my ($node,$grammar) = @_;
  my $treename = treename($node->{tree});
  my $xnode = node($node);
  my $cat = $node->{cat};
  my $url = (defined $grammar && defined $treename) ? "URL=\"$grammar?id=$treename\"" : "";
  my $lemma = $node->{lemma} || '_';
  $treename ||= "";
  my $deriv = 0;
  foreach my $dskel (@{$node->{deriv}}) {
    $deriv += $dskel->count;
  }
  my $label = "$lemma:$cat:$treename ($deriv)";
  $label =~ s/"/\\"/og;
  return <<EOF;
$xnode [label="$label", $url]
EOF
}

sub Forest::collect_dependency {
  my $self = shift;

  ## try to reuse already computed dependencies
  ## ** WARNING ** should check it is enough to store onle $dependency

  if (exists $self->{dependency}) {
    $dependency = $self->{dependency}{dep};
    %trees = %{$self->{dependency}{trees}};
    $forest = $self;
  } else {
    
    $dependency = {};
    $forest = $self;
    %trees = ();
    %derivations = ();
    
    ##  $self->{pseudoop} = {};
    $pseudoop={};
    
    ##  $dcounter="d0000";
    $dcounter="d000000";
    $htid = "ht0001";
    $nid   = "n001";
    $eid   = "e001";

    foreach my $xop (values %{$self->{'op'}}) {
      $xop->dependency();
    }
    
    $self->{dependency} = { dep => $dependency,
			    trees => { %trees }
			  };
  }
}

sub xml_escape {
  my $s = shift;
  $s =~ s/\&/&amp;/og;
  $s =~ s/</&lt;/og;
  $s =~ s/>/&gt;/og;
  $s =~ s/"/&quot;/og;
  return $s;
}

sub Forest::xmldep {
    my $self = shift;

    # First pass: collecting dependencies
    $self->collect_dependency;
##    return;

    my %nodes = ();
    my $GEN   = new MyGenerator('encoding' => 'ISO-8859-1', 
				   ## :pretty = pretty + escaping of illegal XML chars in attributes
				   ## slower than just pretty but useful !
				   ## reverted to pretty + use of own function xml_escape
				  ## 'pretty' => 1 
				  );

    $Forest::XML::Writer::GEN=$GEN;

    my @clusters = ();
    my @nodes    = ();
    my @edges    = ();
    my @op       = ();
    my @ht       = ();
##    my %nodes    = ();
    my %clusters = ();

    my $xmlcluster = sub {
      my ($posa,$posb) = @_;
      my $cid = "c_".$posa."_".$posb;
      return if (exists $clusters{$cid});
      $clusters{$cid} = 1;
      my $cluster = $dependency->{$posa}{$posb};
      $GEN->cluster({ id    => $cid,
		      left  => $posa,
		      right => $posb,
		      lex   => xml_escape($cluster->{lex}) || "",
		      ## deprecated: to be removed in some future
		      token => xml_escape($cluster->{token}) || ""
		    })->emit("  ");
    };

    my $xmlnode = sub {
      my $node = shift;
      my ($posa,$posb)  = @{$node->{pos}};
      my $cat   = $node->{cat};
      my $lemma = $node->{lemma};
      my $tree  = $node->{tree};
      my $token = $node->{token};
      my $id = $node->{nid};
      
      unless (exists $nodes{$id}) {
	$nodes{$id} = 1;
	$xmlcluster->($posa,$posb);
	my $cluster = "c_".$posa."_".$posb;
	my %attr = ( id      =>  $id,
		     cluster => $cluster,
		     tree    => $tree || 'void',
		     cat     => $cat,
		     lemma   => xml_escape($lemma),
		     form => xml_escape($token)
		   );
	my $deriv = $node->{deriv};
	if ($deriv) {
	  $nodes{$id} = $attr{deriv} = join(' ',map {$_->dids} @$deriv);
	}
	$attr{xcat}  = $trees{$tree}{cat} if ($tree);
	$GEN->node(\%attr)->emit("  ");
      }
   };
    
    # Second pass: build xml
    my $edge_handler = 
      sub { my ($self,$target,$edge,$label) = @_;
	    my $source = $edge->{source};
	    $xmlnode->($source);
	    ##			     print STDERR "xml dep edge $ida->$idb label=$label type=$target->{type}\n";
	    my %tmp = ();
	    foreach my $dskel (@{$edge->{deriv}}) {
	      my $info = $dskel->{info};
	      my $dids = $dskel->dids;
	      my $span = join(' ',@{$info->[1]{span}});
	      my $target_op = $info->[1]{id};
	      my $source_op = $info->[0];
	      my $k = "$target_op $source_op";
	      unless (exists $tmp{$k}) {
		$tmp{$k} = { span => $span, 
			     target_op => $target_op,
			     source_op => $source_op,
			     names => "$dids" 
			   };
	      } else {
		$tmp{$k}{names} .= " $dids";
	      }
	    }
	    my @k = keys %tmp;
	    if (scalar(@k) == 1) {
	      ## remove names attr if all derivs from source node go through the edge
	      ## i.e. the edge is a mandatory one !
	      my $k = $k[0];
	      my $v = $tmp{$k};
	      if ($v->{names} eq $nodes{$source->{nid}}) {
		delete $v->{names};
		$v->{mandatory} = 1;
	      }
	    }
	    $GEN->edge({target => $target->{nid},
			source => $source->{nid},
			id     => $eid++,
			type   => $edge->{type},
			label  => $label,
		       },
		       map($GEN->deriv($_),values %tmp)    
		      )->emit("  ");
	  };

    my $node_handler = 
      sub { my ($self,$target) = @_;
	    $xmlnode->($target);
	    $self->edge_loop($target,$edge_handler);
	  };

    my $cluster_handler = 
      sub { my ($self,$posa,$posb,$nodes) = @_;
	    ##				print STDERR "CLUSTER $posa $posb\n";
	    $xmlcluster->($posa,$posb);
	    $self->node_loop( $posa, $posb, $nodes, $node_handler );
	  };

    # Third pass: emit
#    print $GEN->xmldecl;
    my $attrs='';

    $attrs  .= " grammar=\"$self->{grammar}\""  if (defined $self->{grammar});
    $attrs .= " sentence=\"".xml_escape($self->{sentence})."\"" if (defined $self->{sentence});
    $attrs  .= " mode=\"$self->{mode}\"" if (defined $self->{mode});

    print <<EOF;
<?xml version="1.0" encoding="ISO-8859-1" standalone="yes"?>
<dependencies$attrs>
EOF

    $self->cluster_loop( $cluster_handler );

    $self->xmldep_op($GEN);
    $self->xmldep_ht($GEN);

    print <<EOF;
</dependencies>
EOF

    return 1;
  }


sub lp_quote {
  my $string = shift || "";
  return $string if ($string =~ /^[a-z_]\w*$/o);
  return $string if ($string =~ /^\d+$/);
  #    $string =~ s/\'/\'\'/og;
  $string =~ s/\'(?!\')/\'\'/og;
  return "\'$string\'";
}

sub Forest::stats {
    my $self = shift;
    
    # First pass: collecting dependencies
    $self->collect_dependency;

    # Second pass: compute stats
    my $sclusters = 0;
    my $snodes = 0;
    my $sedges = 0;
    my $derivs = 0;

    my $edge_handler = sub { $sedges++; };
    my $node_handler = sub { my ($self,$target) = @_;
			     $snodes++;
			     foreach my $dskel (@{$target->{deriv}}) {
			       $derivs += $dskel->count;
			     }
			     $self->edge_loop( $target, $edge_handler );
			   };
    my $cluster_handler = sub { my ($self,$posa,$posb,$nodes) = @_;
				$sclusters++;
				$self->node_loop( $posa, $posb, $nodes, $node_handler );
			      }; 
    
    $self->cluster_loop( $cluster_handler );
    
    printf "Dependency stats: %2.1f ambiguity %3d clusters %3d nodes %3d edges %d derivs %.1f avgderivs\n",
      (1 + $sedges - $sclusters) / ($sclusters || 1),
	$sclusters,
	  $snodes,
	    $sedges,
	      $derivs,
		$derivs / ($snodes || 1) ;

}

sub Forest::dependency {
    my $self    = shift;
    my $nb      = $self->{'nb'};
    my $grammar = $self->{'grammar'};
    $forest     = $self;

    # First pass: collecting dependencies
    $self->collect_dependency;

    # Second pass: emiting
    if (defined $grammar) {
      print "/* Grammar=$grammar */\n";
    }

    print <<HEADER;
digraph G {
    rankdir = LR;
    ranksep = .5;
    ratio   = auto;
    clusterrank = local;
    node [ style = "filled",  fillcolor = "lightgoldenrod1", fontname  = "sans-serif", fontsize = "10.0" ];
    edge [ fontsize = "10.0", fontname  = "sans-serif" ];
HEADER

    my $lex;
    my $cluster;
    my @trees;
    my $start;

    my $edge_handler = sub { my ($self, $target, $edge, $label) = @_ ;
			     my $source = $edge->{source};
			     my $end    = node($source);
			     my $color  = $color{$edge->{type}};
			     my $weight = abs($source->{pos}[0]-$target->{pos}[0]);
			     $weight    = 1 unless ($weight);
			     my $deriv = 0;
			     foreach my $dskel (@{$edge->{deriv}}) {
			       $deriv += $dskel->count;
			     }
			     my $width = 1;
			     
			     if ($source->{pos}[0] > $target->{pos}[0]) {
			       print <<EDGE;
     $start -> $end [ label = "$label ($deriv)", color = "$color", minlen = $weight, dir = "back", style = "setlinewidth($width)" ];
EDGE
			     } else {
			       print <<EDGE;
     $end -> $start [ label = "$label ($deriv)", color = "$color",  minlen = $weight, style = "setlinewidth($width)" ];
EDGE
			     }
			   };
    my $node_handler = sub { my ($self, $target) = @_;
			     push(@trees, tree_emit($target, $grammar));
			     $start = node($target);
			     $self->edge_loop( $target,  $edge_handler );
			   };
    my $cluster_handler = sub { my ($self,$posa,$posb,$nodes) = @_;
				$cluster = cluster($posa, $posb);
				@trees = ();
				$self->node_loop( $posa, $posb, $nodes, $node_handler );
				## Emit clusters
				my $lex   = $nodes->{'lex'};
				my $trees = join(";\t\n", @trees);
				my $qlex  = $lex;
			##	$qlex = "\\$qlex" if ($qlex eq '"');
				$qlex =~ s/"/\\"/og;
				print <<ANCHOR;
subgraph  $cluster {
	    label     = "$qlex";
            style     = "filled";
            fillcolor = "lightyellow";
	    fontsize  = "10.0";
	    fontname  = "sans-serif";
	    {rank     = same; $trees;}
	}
ANCHOR
			      };

    $self->cluster_loop( $cluster_handler );

    print "}\n\n";
    
  }

sub is_lexical_type {
  my $type = shift;
  return $type eq 'anchor' || $type eq 'lexical' || $type eq 'epsilon';
}

sub Forest::Op::dependency {
    ## returns a list of anchors
    ## and attach a dependency node for anchor ops
    my $self = shift;
    my $tree = shift;
    unless (exists $self->{dep}) {
      my $type = $self->{type};
      if (is_lexical_type($type)) {
	return unless (defined $tree);
	my $cat   = $self->{cat} ||"";
	my $lemma = $self->{lemma} ||"";
	my ($left,$right) = @{$self->{'span'}};
	my $entry =  $dependency->{$left}{$right} ||= {};
	$entry->{lex} ||= $self->{lex};
	## Deprecated: to be removed in some future
	## replaced par 'token/form' on node (rather than token on cluster)
	$entry->{token} ||= $self->{token};
	my $xtree = ($type eq 'epsilon') ? "" : ($type eq 'lexical')  ? "lexical" : $tree;
	## following $e corresponds to a node for $cat $lemma and $xtree
	my $e = $entry->{tree} ||= {};
	$e = $e->{$cat} ||= {};
	$e = $e->{$lemma} ||={};
	$e = $e->{$xtree} 
	  ||= { nid => $nid++,
		cat => $cat,
		lemma => $lemma,
		tree => $xtree,
		token => $self->{token},
		pos => [$left,$right]
	      };
	$self->{dep} = { anchors => [$self], node => $e };
      } else {
	# we return the distinct anchors of the various derivations for op
	## a same anchor may be used for several derivations of a same op
	## => we remove duplicates
	my %anchors = ();
	foreach my $der (@{$self->{'derivations'}}) {
	  my $x = $der->dependency($self);
	  $anchors{$x} = $x;
	}
	$self->{dep} = { anchors => [values %anchors ] };
      }
    }
    return $self->{dep}{anchors};
  }

sub Forest::Deriv::dependency_aux {
    my $self  = shift;
    my $tree = shift;
    my $daux = {};
    my $anchors = $daux->{anchors} = [];
    my $edges = $daux->{edges} = [];
    my $dskel = $daux->{dskel} = [];
    my $lexicals = $daux->{lexicals} = {};
    while (my ($label,$u) = each %{$self->{'nodes'}}){
      foreach my $node (@$u) {
	next unless (defined $forest->{'op'}{$node});
	my $target = $forest->{'op'}{$node};
	if ($target->{'type'} eq 'anchor' 
	    && $target->{'span'}[0] < $target->{'span'}[1] ){
	  push(@$anchors,$target);
	  $trees{$tree}{node} = $label;
	  next;
	}
	$lexicals->{$label} = $target if ($target->{type} eq 'lexical');
	if ($label =~ /^\d+/) {
	  my $cat =  $target->{'cat'};
	  $label = "$cat\#$label";
	}
	my $tanchors = $target->dependency($tree);
	push(@$edges, { 'type'       => $target->{'type'},
			'label'      => $label,
			'anchors'    => $tanchors,
			'targetspan' => $target->{span}, # this one should be removed in favor of target
			'target'     => $target
		      } 
	    );
	if (scalar(@$tanchors) > 1) {
	  my $i=1;
	  push(@$dskel,{ map {$_ => $i++} @$tanchors });
	}
      }
    }
    return $daux;
  }

sub Forest::Deriv::dependency {
    # return an anchor
    my $self  = shift;
    my $xop    = shift;
    my $opid = $xop->{id};
    my $tree  = $self->{tree};
    my $ht = $self->{ht};
    my $did   = $self->{id} ||= $dcounter++;
    $trees{$tree} = { cat   => $xop->{cat} , 
		      lemma => $xop->{lemma} || ""
		    };

    my $daux = $self->dependency_aux($tree);

    my $lexicals = $daux->{lexicals};
    my $edges = $daux->{edges};
    my $anchors = $daux->{anchors};
    
    ## WARNING: we may have cases of derivation with no anchor and no indication of tree
    ## what should we do ? maybe use the anchor of some node in deriv as anchor
    ##    print "Deriv for $opid $xop->{cat} @{$xop->{span}} $self->{tree}: anchor=@$anchors edges=@$edges\n";
    unless (@$anchors) {
      ## We create a pseudo anchor,
      my $xid = "$tree @{$xop->{span}}";
      my $dummy = $forest->{pseudoop}{$xid}
	||= Forest::Op->new( id    => $xid,
			     type  => 'anchor',
			     lex   => "",
			     token => "",
			     lemma => "",
			     deriv => {},
			     tree  => $tree,
			     span  => [$xop->{span}[0],$xop->{span}[0]],
			     cat   => $xop->{cat} 
			   );
      $dummy->dependency($tree);
      push(@$anchors,$dummy);
    }
    my $anchor = $anchors->[0];
    $anchor->dependency($tree);
    ## $sanchor correspond to the start node
    my $sanchor = $anchor->{dep}{node};
    
    # Skeletons of all possible derivation indexes
    # when ambiguity for the head of the target of a dependency, we have to duplicate the derivation id
    my $dskel = $daux->{dskel};
    
    ##    my @dids = map("$did$_",gen_did(0,$dskel));
    my $AllDerivs = DerivSkel->new({ prefix => "$did",
				     filter => 0,
				     dskel => $dskel
				   });
    push(@{$sanchor->{deriv}}, $AllDerivs);
    $self->{dids} = $AllDerivs;
    
    foreach my $edge (@$edges) {
      ## for each edge, we store a dep link $info
      ## from each target anchors
      ## to the source anchor (modulo some rerouting for adj on lexical coanchors)
      ## To each dep link, we build and maintain
      ## the list of derivs going through this link
      my $label = $edge->{label};
      my $type = $edge->{type};
      my $amb = scalar(@{$edge->{anchors}}) > 1;
      foreach my $start (@{$edge->{anchors}}) {
	$start->dependency();
	my $reroot = 0;
	## To handle operations on lexical nodes that should be attached to them
	## rather than to the anchor of the tree
	my $source = $anchor;
	my $stree  = $start->{tree} || "";
	## the tree of a lexical is the same tree of the anchor
	$stree = 'lexical' if ($type eq 'lexical');
	if (exists $lexicals->{$label} && $type eq 'adj') {
	  $source = $lexicals->{$label};
	  $reroot = 1;
	  ## WARNING: This rerooting of adj on lexical nodes is not yet complete
	}
	my $sdep = $start->{dep}{node};
	my $sedges = $sdep->{label}{$label} ||= [];
	my $ops_info  = [$opid,$edge->{target}];
	my $eskel = DerivSkel->new({ filter => $start,
				     prefix => $did,
				     dskel => $dskel,
				     info => $ops_info
				   }
				  );
	$eskel->{like} = $AllDerivs unless ($amb);
	my $info = { type  => $type,
		     source => $source->{dep}{node},
		     deriv => [ $eskel ],
		     reroot => $reroot
		   };
	next if (grep {is_edge($_,$info)} @$sedges);
	push(@$sedges,$info);
      }
    }
    return $anchor;
  }

sub is_edge {
    my $e1 = shift;
    my $e2 = shift;
    if ( $e1->{'type'} eq $e2->{'type'} 
	 && $e1->{source} == $e2->{source}
       ) {
      ## Complete set of derivation using e1 with those using e2
      push(@{$e1->{deriv}},@{$e2->{deriv}});
      return 1;
    } else {
      return 0;
    }
}

sub norm_tree {
    my $tree = shift || "";
#    print "/* norm tree $tree */\n";
    $tree =~ tr/-/_/;
    $tree =~ s/\s+/_/og;
    return $tree;
}

sub Forest::xmldep_ht {
  my $self = shift;
  my $GEN  = shift;

  foreach my $head (keys %{$self->{'op'}}) {
    $self->{'op'}{$head}->xmldep_ht($GEN);
  }

}

sub Forest::Op::xmldep_ht {
  my $self  = shift;
  my $GEN   = shift;
  my $type  = $self->{type};
  my $id    = $self->{'id'};
  return if ($type eq 'anchor' || $type eq 'lexical');
 
  foreach (@{ $self->{'derivations'}}) {
    $_->xmldep_ht($GEN,$id);
  }
}

sub Forest::Deriv::xmldep_ht {
  my $self = shift;
  my $GEN   = shift;
  my $opid = shift;
  my $ht = $self->{ht};
  if (defined $ht) {
    $GEN->hypertag({ derivs => $self->{dids}->dids,
		     op => $opid,
		     id => $htid++
		   },
		   $ht->pretty_print(2)
		  )->emit("  ");
  }
}

sub Forest::xmldep_op {
  my $self = shift;
  my $GEN  = shift;

  foreach my $head (keys %{$self->{'op'}}) {
    $self->{'op'}{$head}->xmldep_op($GEN);
  }

}

sub Forest::Op::xmldep_op {
  my $self  = shift;
  my $GEN   = shift;
  my $id    = $self->{'id'};
  my $type  = $self->{type};
  return if ($type eq 'anchor' || $type eq 'lexical');
  my @deriv = ();
  my $derivs = join(' ',map {$_->xmldep_op($GEN,$id)} @{$self->{derivations}});

  my @content=();
  push(@content,$self->{'top'}->pretty_print(2,'top')) if (defined $self->{'top'});
  push(@content,$self->{'bot'}->pretty_print(2,'bot')) if (defined $self->{'bot'});
  $GEN->op({id => $id, span => join(' ', @{$self->{span}}), deriv => $derivs, cat => $self->{cat}}, @content)->emit("  ");
}

sub Forest::Deriv::xmldep_op {
  my $self = shift;
  return $self->{dids}->dids;
}

package DerivSkel;
#our @ISA = qw/DerivSkel/;

sub new {
  my $this = shift;
  my $class  = ref($this) || $this;
  my $self = shift;
  bless $self, $class;
  return $self;
}

sub dids {
  my $self = shift;
  return $self->{memodids} if (exists $self->{memodids});
  return $self->{like}->dids if (exists $self->{like});
  my $prefix = $self->{prefix};
  my $dskel = $self->{dskel};
  my $filter = $self->{filter};
##  print "Handling dids $prefix @$dskel\n";
  my $dids = join(' ', map {"$prefix$_"} gen_did($filter,$dskel) );
  $self->{memodids} = $dids;
  return $dids;
}

sub gen_did {
  my $filter  = shift;
  my $dskel    = shift;
  return "" unless (@$dskel);
  my @dids = ("");
  foreach my $anchors (@$dskel) {
    if (my $l = $anchors->{$filter}) {
      @dids = map($_."_$l",@dids);
    } else {
##      my $l = join("|",values %$anchors);
      @dids = map {my $v=$_; map {$_."_$v"} @dids} (values %{$anchors});
##      @dids = map($_."_$l",@dids);
    }
  }
  return @dids;
}

sub count {
  my $self = shift;
  return $self->{memocount} if (exists $self->{memocount});
  return $self->{like}->count if (exists $self->{like});
  my $dskel = $self->{dskel};
  my $filter = $self->{filter};
  my $n = 1;
  foreach my $a (@$dskel) {
    next if (defined $a->{$filter});
    $n *= scalar(keys %$a);
  }
  $self->{memocount} = $n;
  return $n;
}


package MyGenerator;

## A very specific XML generator to be much faster than using XML::Generator

sub new {
  my $class = shift;
  return {}, $class;
}

sub node {
  my $self = shift;
  my $a = shift;
  return MyGenerator::Elt->new("node",$a);
}

sub edge {
  my $self = shift;
  my $a = shift;
  return MyGenerator::Elt->new("edge",$a,@_);
}

sub deriv {
  my $self = shift;
  my $a = shift;
  return simple("deriv",$a);
}

sub op {
  my $self = shift;
  my $a = shift;
  return MyGenerator::Elt->new("op",$a,@_);
}

sub cluster {
  my $self = shift;
  my $a =shift;
  return MyGenerator::Elt->new("cluster",$a);
}

sub dependencies {
  my $self = shift;
  my $a={};
  $a = shift if (ref($_[0]) eq "HASH");
  return MyGenerator::Elt->new("dependencies",$a,@_);
}

sub narg {
  my $self = shift;
  my $a=shift;
  return MyGenerator::Elt->new("narg",$a,@_);
}

sub hypertag {
  my $self = shift;
  my $a = shift;
  return MyGenerator::Elt->new("hypertag",$a,@_);
}

sub fs {
  my $self = shift;
  my $a={};
  $a = shift if (ref($_[0]) eq "HASH");
  return MyGenerator::Elt->new("fs",$a,@_);
}

sub f {
  my $self = shift;
  my $a={};
  $a = shift if (ref($_[0]) eq "HASH");
  return MyGenerator::Elt->new("f",$a,@_);
}

sub minus {
  return "<minus/>";
}

sub plus {
  return "<plus/>"
}

sub val {
  my $self = shift;
  my $val = shift;
  return "<val>$val</val>";
}

sub simple {
  my ($name,$attrs) = @_;
  my $a = '';
  while (my ($k,$v) = each %$attrs) {
    next unless (defined $v);
    $a .= " $k=\"$v\"";
  }
  return "<$name$a/>";
}

package MyGenerator::Elt;

sub new {
  my ($class,$name,$attrs,@content) = @_;
  return bless { name => $name, 
		 attrs => $attrs,
		 content => [@content]
	       }, $class;
}

sub emit {
  my $self = shift;
  my $pref = shift;
  my $txt = "$pref<$self->{name}";
  while (my ($k,$v) = each %{$self->{attrs}}) {
    next unless (defined $v);
    $txt .= " $k=\"$v\"";
  }
  if (@{$self->{content}}) {
    $txt .= ">\n";
    print $txt;
    my $spref = "$pref ";
    foreach (@{$self->{content}}) {
      if (ref($_)) {
	$_->emit($spref);
      } else {
	print "$spref$_\n";
      }
    }
    print "$pref","</$self->{name}>\n";
  } else {
    $txt .= "/>\n";
    print $txt;
  }
  delete $self->{content};
  delete $self->{attrs};
}

1;

__END__

=head1 NAME

Forest::Dependency::Writer

=head1 DESCRIPTION

Writes forest of dependencies in dot format to build a graph.

=head1 SYNOPSIS

  use Forest::Dependency::Writer;

=head1 COPYRIGHT AND LICENSE

Copyright (C) 2000-2007, INRIA.

This program is free software; you can redistribute it and/or modify it under
the same terms as Perl itself.

=head1 AUTHOR

Eric de la Clergerie <Eric.De_La_Clergerie@inria.fr>

=cut

