# $Id: Parser.pm 2059 2008-08-27 16:38:32Z clerger $
package Forest::LP::Parser;

require 5.005;
use strict;
use warnings;
use Forest;

## use Data::Dumper;

#our $pattern = qr/(\w+)\s*(?:{(.*?)})?\(\s*(\d+)\s*,\s*(\d+)\s*\)/o;
our $pattern = qr/(\w+)\s*(?:{(.*?)})?\(\s*(\S+)\s*,\s*(\S+)\s*\)/o;
our $forest;
our %vars;
our %shared=();
our %bodies=();

our $type = 'lp';

our $config = { 'lp' => { 'adj' => \&lp_adj_parse,
			  'std' => \&lp_std_parse,
			  'anchor' => \&anchor_parse,
			  'lexical' => \&lexical_parse,
			  'struct' => \&struct_parse,
			  'epsilon' => \&epsilon_parse,
			},
		'xrcg' => { 'adj' => \&xrcg_adj_parse,
			    'std' => \&xrcg_std_parse,
			    'anchor' => \&anchor_parse,
			    'lexical' => \&lexical_parse
			    }
	    };

our $anonkey = "anon000000";
our $bodyid = 0;

#$::RD_TRACE=1;
$::RD_AUTOACTION = q{ $item[1] };

our $redid = "a000000";
our %reduce = ();

sub reduce_register_and_emit {
  my $s = shift;
##  print STDERR "try reduce '$s'\n";
  if (exists $reduce{$s}) {
    return "#".$reduce{$s};
  } else {
    my $shared = $reduce{$s} = $redid++;
    my $current = body_parse($s);
    $shared{"shared_$shared"} 
      = Forest::Shared->new(key=> "shared_$shared",
			    shared => $current->{body},
			    id => $current->{id},
			    sign => $current->{sign}
			   );
    return "#".$shared;
  }
}

sub production_parse {
  my $prod = shift;
  my ($head,$body) = ($prod =~ /(\d+)\s*<--\s*(.*)\s*$/);

  $body =~ s/^\((.*)\)\s*$/$1/o;

#  print "Body $body\n";

  ## progressively flatten the body by introducing shared/indirect subparts
  while($body =~ /\(/) {
    $body =~ s/\(\s*([^()]+?)\s*\)/reduce_register_and_emit($1)/eg;
  }

  #  print STDERR "Reduced $body\n";

  $body = body_parse($body);
  return ($head,$body->{body});
}

sub body_parse {
  my $body = shift;
  #  print STDERR "body parse $body\n";
  my @stack = ();
  my $alternatives = [];
  my $current = [];
  my $shared;

 LOOP: {
    last LOOP if ($body =~ /\G\s*$/ogc);
    redo LOOP if ($body =~ /\G\s+/ogc);
    push(@$current,["$1" => "$2"]), redo LOOP if ($body =~ /\G\[\s*([\w<>_\d!-]+)\s*\]\s*(\d+)\s*\s*/ogc);
    push(@$current,['void' => "$1"]), redo LOOP if ($body =~ /\G(\d+)\s*/ogc);
    if ($body =~ /\G\|\s*/ogc) {
#      print STDERR "start alternative\n";
      push(@$alternatives,$current);
      $current = [];
      redo LOOP;
    }
    if ($body =~ /\G\(\s*/ogc) {	
#      print STDERR "open body\n";
      unshift(@stack,[$current,$alternatives,$shared]);
      $current = [];
      $alternatives = [];
      $shared = "shared_".$anonkey++;
      redo LOOP;
    }
    if ($body =~ /\G\#(a?\d+)\s*\{\s*/ogc) {
#      print STDERR "open shared body\n";
      unshift(@stack,[$current,$alternatives,$shared]);
      $current = [];
      $alternatives = [];
      $shared = "shared_$1";
      redo LOOP;
    }
    if ($body =~ /\G\#(a?\d+)\s*/ogc) {
#      print STDERR "reusing shared body\n";
      my $tmp = $shared{"shared_$1"} ||= Forest::Shared->new(key=>"shared_$1");
      push(@$current,[shared => $tmp]);
      redo LOOP;
    }
    if ($body =~ /\G[)}]\s*/ogc) {
#      print STDERR "closing (shared) body\n";
      push(@$alternatives,$current) if (@$current || @$alternatives);
      $current = sign_and_register($alternatives);
      #     print STDERR "built body id=>$current->{id} sign=>$current->{sign}\n";
      warn "WARNING: undefined shared key" unless (defined $shared);
      my $tmp =  $Forest::LP::Parser::shared{$shared};
      warn "WARNING: shared key $shared already defined" if (defined $tmp);
      $tmp = $shared{$shared} ||= Forest::Shared->new(key=>$shared);
      $tmp->{shared} = $current->{body};
      $tmp->{id} = $current->{id};
      $tmp->{sign} = $current->{sign};
      warn "*** ERROR Expected non empty stack" unless (@stack);
      my $saved = shift @stack;
      my $label = $shared;
      ($current,$alternatives,$shared) = @$saved;
      push(@$current, [ $label => $tmp]);
      redo LOOP;
    }
    warn "WARNING: should not be here\n";
  }
  warn "*** ERROR Expected empty stack" if (@stack);
  # don't forget to build whole body
  push(@$alternatives,$current) if (@$current || @$alternatives);
  $current = sign_and_register($alternatives);
  return $current;
}

# $::RD_HINT=1;

sub fsargs_parse {
  my $fs = shift;
  my @stack;
  my $current = Forest::Feature->new( f=>{});	# current list of args
  my $comp;			# current feature value
  pos($fs)=0;
 LOOP: {
#    print STDERR "next iteration pos=".pos($fs)."\n";
    last LOOP if $fs =~ /\G\s*$/ogc;
    redo LOOP if $fs =~ /\G\s+/ogc;
    if ($fs =~ /\G([\w\d-]+)\{/ogc) {
#      print STDERR "opening fs $1\n";
      unshift(@stack,[$current,$comp]);
      $current = Forest::Feature->new(type => $1, f => {} );
      $comp->{value} = [$current];
      redo LOOP;
    }
    if ($fs =~ /\G([\w\d-]+)\s*=>\s*/ogc) {
#      print STDERR "reading feature $1\n";
      $comp = Forest::Comp->new( comp=> $1, value => [] );
      $current->{f}{$1} = $comp;
      if ($fs =~ /\G([A-Z_][\w\d]*)\s*((?:::|,)?)\s*/ogc){
#	print STDERR "reading variable $1\n";
	$comp->{$2 eq "::" ? 'var' : 'id'} = $1;
      };
      redo LOOP;
    }
    if ($fs =~ /\G([\w\d-]+)\[(.*?)\]\s*(,\s*)?/ogc) {
#      print STDERR "reading fset value $1 [$2]\n";
      $comp->{value} = [map(fs_value($_),split(/\s*,\s*/,$2))];
      redo LOOP;
    }
    if ($fs =~ /\G(\+|\-|[^,\[\]\{\}() ]+)\s*(,\s*)?/ogc) {
#      print STDERR "reading atomic value $1\n";
      $comp->{value} = [fs_value($1)];
      redo LOOP;
    }
    if ($fs =~ /\G\}\s*(,\s*)?/ogc) {
#      print STDERR "closing fs\n";
      my $tmp = shift @stack;
      ($current,$comp) = @$tmp;
      redo LOOP;
    }
    my $pos = pos($fs);
    $fs =~ /\G(.*)/ogc;
    warn "WARNING: fsarg parser: should not arrive here at pos $pos: $fs\n\t at '$1'";
  }
  return $current;
}

sub fs_value {
  my $v = shift;
  return Forest::Plus->new() if ($v eq '+');
  return Forest::Minus->new() if ($v eq '-');
  return Forest::Val->new( val => $v );
}

sub lp_adj_parse {
    my $nt = shift;
    if ($nt =~ /^$pattern\s*\*\s*$pattern/o) {
	return { 'cat' => $1, 'top' => $2, 'bot' => $6, 'span' => [$3,$4,$7,$8] };
    } else {
	return 0;
    }
}

sub lp_std_parse {
    my $nt = shift;
    if ($nt =~ /^$pattern/o) {
	return { 'cat' => $1, 'top' => $2, 'span' => [$3,$4] };
    } else  {
	return 0;
    }
  }

sub struct_parse {
    my $nt = shift;
    if ($nt =~ m/^verbose!struct\(
                        \s*(.+?)\s*,   # Tree
			\s*(.+?)\s*
		 \)\s*$/ox) {
      ## New since July 2008: adding HyperTag to struct
      my ($tree,$ht) = ($1,$2);
      $ht =~ s/ht{(.*)}/$1/o;
##      print "HT $ht\n";
      return { 'tree' => $tree, 
	       'ht' => $ht
	     };
    } elsif ($nt =~ m/^verbose!struct\(
                        \s*(.+?)\s*   # Tree
		 \)/ox) {
      return { 'tree' => $1 };
    } else {
      return 0;
    }
  }

sub anchor_parse {
    my $nt = shift;
    if ($nt =~ m/^verbose!anchor\(
                        \s*(.+?)\s*,    # Token
                        \s*(\S+?)\s*,    # Left
                        \s*(\S+?)\s*,    # Right
                        \s*(.+?)\s*,    # Tree
                        \s*(\w+)\s*(?:{(.*)})?\s*,    # Cat{...}
                        \s*\[(.+?),(.+?)\]\s*     # Lex
                        (,\s*tag_anchor{.*}\s*)?
                 \)/ox) {
      return { 'lex' => $8, token => $1, 'lemma' => $7, 'cat' => $5, 'bot' => $6, 'span' => [$2,$3], 'tree' => $4 };
    } elsif ($nt =~ m/^verbose!anchor\(
                        \s*(.+?)\s*,    # Token
                        \s*(\S+?)\s*,    # Left
                        \s*(\S+?)\s*,    # Right
                        \s*(.+?)\s*,    # Tree
                        \s*(\w+)\s*(?:{(.*)})?\s*,    # Cat{...}
                        \s*(.+?)\s*     # Lex
                 \)/ox) {
      # soon to be deprecated : May 11 2005
      return { 'lex' => $7, token => $1, 'cat' => $5, 'bot' => $6, 'span' => [$2,$3], 'tree' => $4 };
    } elsif ($nt =~ /^tag_anchor\(\s*(.+?)\s*,\s*(\S+?)\s*,\s*(\S+?)\s*,\s*(.+?)\s*\)/o) {
      # Deprecated
      return { 'lex' => $1, token => $1, 'span' => [$2,$3], 'tree' => $4 };
    } else {
      return 0;
    }
}

sub lexical_parse {
    my $nt = shift;
    if ($nt =~ m/^verbose!lexical\(
                         \s*\[\s*(.*?)\s*\]\s*,  ## Token
                         \s*(\S+?)\s*,            ## Left
                         \s*(\S+?)\s*,            ## Right
                         \s*(.+?)\s*,            ## Cat
                         \s*\[(.+?),(.+?)\]\s*             ## Lex
                 \)/xo) {
      return { 'lex' => $6, token => $1, 'lemma' => $5, 'bot' => '', 'span' => [$2,$3], cat => $4 };
    } elsif ($nt =~ m/^verbose!lexical\(
                         \s*\[\s*(.*?)\s*\]\s*,  ## Token
                         \s*(\S+?)\s*,            ## Left
                         \s*(\S+?)\s*,            ## Right
                         \s*(.+?)\s*,            ## Cat
                         \s*(.+?)\s*             ## Lex
                 \)/xo) {
      # soon to be deprecated : May 11 2005
      return { 'lex' => $5, token => $1, 'bot' => '', 'span' => [$2,$3], cat => $4 };
    }  elsif ($nt =~ m/^verbose!coanchor\(
		       \s*(.+?)\s*,    # Token
		       \s*(\S+?)\s*,    # Left
		       \s*(\S+?)\s*,    # Right
		       \s*(\w+)\s*(?:{(.*)})?\s*,    # Cat{...}
		       \s*\[(.+?),(.+?)\]\s*     # Lex
                 \)/ox) {
      return { 'lex' => $7, token => $1, 'lemma' => $6, 'cat' => $4, 'bot' => $5, 'span' => [$2,$3] };
    } elsif ($nt =~ m/^verbose!coanchor\(
                        \s*(.+?)\s*,    # Token
                        \s*(\S+?)\s*,    # Left
                        \s*(\S+?)\s*,    # Right
                        \s*(\w+)\s*(?:{(.*)})?\s*,    # Cat{...}
                        \s*(.+?)\s*     # Lex
                 \)/ox) {
      # soon to be deprecated : May 11 2005
      return { 'lex' => $6, token => $1, 'cat' => $4, 'bot' => $5, 'span' => [$2,$3] };
    } elsif ($nt =~ /^tag_lexical\(\s*\[\s*(.*?)\s*\]\s*,\s*(\S+?)\s*,\s*(\S+?)\s*\)/o) {
      # Deprecated
      return { 'lex' => $1, token => $1, 'span' => [$2,$3] };
    } else {
      return 0;
    }
}

sub epsilon_parse {
  my $nt = shift;
  if ($nt =~ m/^verbose!epsilon\(
	       \s*(.+?)\s*,             ## Token
	       \s*(\S+?)\s*,            ## Left
	       \s*(\S+?)\s*,            ## Right
	       \s*\[(.+?),(.+?)\]\s*    ## Lex
	       \)/xo) {
    ##    warn "FOUND EPSILON\n";
    return { 'lex' => $5, token => $1, 'lemma' => $4, 'span' => [$2,$3] };
  } else {
    return 0;
  }
}

our $xrcg_pair = qr/\((\S+)\s*:\s*(\S+)\)/o;

sub xrcg_adj_parse {
    my $nt = shift;
    if ($nt =~ /^adj_(\w+)\(\w+(?:{(.*?)})?,\w+(?:{(.*?)})?\)\s*\($xrcg_pair,$xrcg_pair\)/o) {
	return { 'cat' => $1, 'top' => $2, 'bot' => $3, 'span' => [$4,$6,$5,$7] };
    } else {
	return 0;
    }
}

sub xrcg_std_parse {
    my $nt = shift;
    if ($nt =~ /^(\w+)(?:{(.*?)})?\s*\($xrcg_pair\)/o) {
	return { 'cat' => $1, 'top' => $2, 'span' => [$3,$4] };
    } else  {
	return 0;
    }
}

sub parse {
    my $handle = shift;
    my $options = shift || {};
    my $keep = $options->{keep};
    $forest = Forest->new( {@_, 'op'=>{} } );
    %shared = ();
    $anonkey = "anon000000";
    %bodies = ();
    $bodyid = 0;
    $redid = "a000000";
    %reduce = ();

    $forest->{'nb'} = -1;

    while (<$handle>){
      if (/^\*+.*robust\s+mode/i) { # forest issued by partial robust parsing
	$forest->{mode} = 'robust';
	next;
      }
      last if (/^Shared Forest/);
    }

    while (<$handle>){
#	print "LINE $_";
	next if (/^\s*$/);	# skip blank lines
	next if (/^Time\s+\d+/); # skip Time info

	if (/^\*ANSWER\*/) {	# skip ANSWER production
	    $_=<$handle>;
	    next;
	}
	s/__(\d+)/$1/og;		# Remove trailing number for variables
	my $nt=$_;
	my $production = <$handle>;
	chop $nt;
	chop $production;
	%vars=();
	my $info;
#    print "Dealing with $nt\n\t$production\n";
	my ($head,$derivations) = production_parse($production);
	if ($info=&{$config->{$type}{'adj'}}($nt)) {
	    # Adjunctions
	    &var_set($info->{'top'});
	    &var_set($info->{'bot'});
	    $forest->{'op'}{$head} = Forest::Op->new( 'id' => $head,
						      'type' => 'adj',
						      'derivations' => $derivations,
						      'span' => $info->{'span'},
						      'cat' => $info->{'cat'},
						      'top' => &handle_avm($info->{'top'}),
						      'bot' => &handle_avm($info->{'bot'})
						  );
	} elsif ($info=&{$config->{$type}{'std'}}($nt)) {
	    # Substitutions
	    &var_set($info->{'top'});
	    $forest->{'op'}{$head} = Forest::Op->new( 'id' => $head,
						      'type' => 'subst',
						      'derivations' => $derivations,
						      'span' => $info->{'span'},
						      'cat' => $info->{'cat'},
						      'top' => &handle_avm($info->{'top'})
						      );
	} elsif ($info=&{$config->{$type}{'anchor'}}($nt)) {
	    # Anchors
	    &var_set($info->{'bot'});
	    $forest->{'op'}{$head} = Forest::Op->new( 'id' => $head,
						      'type' => 'anchor',
						      'derivations' => $derivations,
						      'span' => $info->{'span'},
						      'lex' =>$info->{'lex'} || "",
						      'bot' => &handle_avm($info->{'bot'}),
						      'cat' =>$info->{'cat'} || "",
						      'tree' => $info->{'tree'},
						      lemma => $info->{lemma} || "",
						      token => $info->{token} || ""
						      );
	} elsif ($info=&{$config->{$type}{'lexical'}}($nt)) {
	    # Lexicals
	    &var_set($info->{'bot'});
	    $forest->{'op'}{$head} = Forest::Op->new( 'id' => $head,
						      'type' => 'lexical',
						      'derivations' => $derivations,
						      'span' => $info->{'span'},
						      'lex' =>$info->{'lex'} || "",
						      'cat' =>$info->{'cat'} || "",
						      'bot' => &handle_avm($info->{'bot'}),
						      lemma => $info->{lemma} || "",
						      token => $info->{token} || ""
						    );
	  } elsif ($info=&{$config->{$type}{'struct'}}($nt)) {
	    # Structs
	    my $ht = $info->{ht};
	    if (defined $ht) {
	      &var_set($ht);
	      $ht = &handle_avm($ht);
	      $ht->{type} = 'ht';
	    }
	    $forest->{'op'}{$head} = Forest::Op->new( 'id' => $head,
						      'type' => 'struct',
						      'derivations' => $derivations,
						      'tree' => $info->{'tree'},
						      'ht' => $ht
						    );
	  } elsif ($nt =~ /^void_adj/) {
	    # deprecated old format
	  } elsif ($info=&{$config->{$type}{'epsilon'}}($nt)){
	    $forest->{'op'}{$head} = Forest::Op->new( id => $head,
						      type => 'epsilon',
						      derivations => $derivations,
						      span => $info->{'span'},
						      lex =>$info->{'lex'} || "",
						      lemma => $info->{lemma} || "",
						      token => $info->{token} || ""
						    );
	} else {
##	    warn "Not a correct non terminal $nt";
	}
	$forest->{'nb'}=$head if ($head > $forest->{'nb'}) ;
    }
    ## A phrase of normalization
    ## To assign to each derivation its associated tree
    my @delete = ();
    foreach my $id (keys %{$forest->{'op'}}) {
      my $op = $forest->{'op'}{$id};
      my $type = $op->{type};
      if ($type eq 'struct') {
	push(@delete,$id);
	next;
      }
      foreach my $d (@{$op->{derivations}}) {
	$d->decore($forest);
      }
    }
    delete $forest->{op}{$_} foreach (@delete);
    ## to filter out some long derivations
    if (defined $keep) {
      foreach my $op (values %{$forest->{op}}) {
	my $derivs = $op->{derivations};
	if (scalar(@$derivs) > 1) {
	  my @tmp = grep {$_->keep($keep)} @$derivs;
	  if (@tmp && (scalar(@tmp) < scalar(@$derivs))) {
	    $op->{derivations} = [@tmp];
	    $op->{all_derivations} = $derivs;
	  }
	}
      }
    }
    return $forest;
}

sub Forest::Deriv::keep {
  my $self = shift;
  my $keep = shift;
  my $n = 0;
  foreach my $l (values %{$self->{nodes}}) {
    my $m = 0;
    foreach my $opid (@$l) {
      my $op = $forest->{op}{$opid};
      next unless (defined $op);
      $m++ if ($op->{type} eq 'adj');
    }
    return 0 if ($m > $keep);
    $n += $m;
  }
  return ($n < 20);
}


sub Forest::Deriv::decore {
  my $self = shift;
  my $forest = shift;
  unless ($self->{decorated}) {
    $self->{decorated} = 1;
    foreach my $u (values %{$self->{nodes}}) {
      foreach my $v (@$u) {
	if (ref($v) eq "Forest::Shared") {
	  ## shared subbody
	  $v->decore($forest);
	  $self->{tree} = $v->{tree} if (defined $v->{tree});
	  $self->{ht} = $v->{ht} if (defined $v->{ht});
	} else {
	  my $subop = $forest->{'op'}{$v};
	  $self->{tree} = $subop->{tree} if (defined $subop->{tree});
	  $self->{ht} = $subop->{ht} if (defined $subop->{ht});
	}
      }
    }
  }
}

sub Forest::Shared::decore {
  my $self = shift;
  my $forest = shift;
  unless ($self->{decorated}) {
    $self->{decorated} = 1;
    foreach my $d (@{$self->{shared}}) {
      $d->decore($forest);
      $self->{tree} = $d->{tree} if (defined $d->{tree});
      $self->{ht} = $d->{ht} if (defined $d->{ht});
    }
  }
}

sub deriv_signature {
  my $d = shift;
  my @s =();
  foreach my $key (sort keys %$d) {
    push(@s," ++$key", sort {$a <=> $b} @{$d->{$key}});
  }
  return join(" ",@s);
}

sub display_deriv {
  my $deriv = shift;
##  print STDERR "DERIV ";
  foreach my $key (sort keys %{$deriv->{nodes}}) {
    print STDERR "$key:{@{$deriv->{nodes}{$key}}} "
  }
##  print STDERR "\n";
}

sub sign_and_register {
  ## $body is an array (alternatives) of arrays (derivs) of either
  ##      - pairs [label => nt]
  ##      - shared parts [shared => Forest::Shared({key => <key>, id => <id>, shared => <normbody>})
  my $body = shift;
  my %tmp = ();
  foreach my $deriv (@$body) {
    my @deriv = @$deriv;
    my @s1 = ();
    while (@deriv) {
      my $lnt = shift @deriv;
      my ($label,$node) = @$lnt;
      if ($label =~ /^shared_/) {
	push(@s1,'#'.$node->{id});
      } else {
	push(@s1,"$label.$node");
      }
    }
    my $sig = join(" ",sort @s1);
    $tmp{$sig} = $deriv;
  }
  my $s = join(' | ',sort keys %tmp);
  my $e = $bodies{$s} 
    ||= { sign => $s, 
	  id => ++$bodyid, 
	  body => body_normalize([values %tmp])
	};
  return $e;
}

sub body_normalize {
  my $body = shift;
  my $all = [];
  my $alls = {};
  foreach my $deriv (@$body) {
    my @allderivs = ({});
    my @deriv = @$deriv;
    while (@deriv) {
      my $lnt = shift @deriv;
      my ($label,$node) = @$lnt;
      my $s = {};
##      print STDERR "handling label $label\n";
      $s = $alls unless (@deriv);
      if (ref($node) eq "Forest::Shared") {
	my $shared = $node->{shared}; # an array of normalized derivs
	my @tmpderivs = ();
	foreach my $tmp (@$shared) {
	  my $sderiv = $tmp->{nodes};
	  push(@tmpderivs,map {deriv_add($_,$sderiv,$s)} @allderivs );
	}
	@allderivs = @tmpderivs;
      } else {
	@allderivs = map { deriv_add($_,{$label => [$node]},$s) } @allderivs;
      }
    }
    push(@$all,map(Forest::Deriv->new(nodes=>$_),@allderivs));
  }
  return $all;
}


sub deriv_add {
##  print STDERR "deriv add\n";
  my $d1 = shift;
  my $d2 = shift;
  my $s = shift;
  my $new = { %$d1 };
  foreach my $l (keys %$d2) {
    if (exists $new->{$l}) {
      ## beware of copying from $d1
      $new->{$l} = [@{$d1->{$l}},@{$d2->{$l}}];
    } else {
      ## reuse from $d2
      $new->{$l} = $d2->{$l};
    }
  }
  my $sig = deriv_signature($new);
##  print STDERR "testing sign $sig\n";
  return if (exists $s->{$sig});
  $s->{$sig} = 1;
##  print STDERR "adding deriv $sig\n";
  return $new;
}

sub handle_avm {
  my $avm = shift || '';
#  return $fsparser->args($avm);
  return fsargs_parse($avm);
}

sub handle_value {
    shift;
    s/^\s*//o;
    s/\s*$//;
    return Forest::Plus->new() if /^\+$/;
    return Forest::Minus->new() if /^-$/;
    return Forest::Val->new( val => $_ );
}

sub var_set {
  my $p = shift;
  return unless $p;
  $vars{$_}++ foreach ($p =~ /[A-Z_]\w*/og);
}

package Forest::LP::XRCG::Parser;

sub parse {
    $Forest::LP::Parser::type='xrcg';
  my $forest = Forest::LP::Parser::parse(@_);
    $Forest::LP::Parser::type='lp';
    return $forest;
}

1;
__END__

=head1 NAME

Forest::LP::Parser - Parse the LP representation of TAG Derivation forest

=head1 SYNOPSIS

  use Forest::LP::Parser;

=head1 DESCRIPTION

=head2 EXPORT

None by default.

=head1 SEE ALSO

Forest (3)
perl(1).

=head1 COPYRIGHT AND LICENSE

Copyright (C) 2000-2007, INRIA.

This program is free software; you can redistribute it and/or modify it under
the same terms as Perl itself.

=head1 AUTHOR

Eric de la Clergerie <Eric.De_La_Clergerie@inria.fr>

=cut


sub handle_avm {
    my %fs = ();
    my $avm = shift || '';
    $avm =~ /^/g;
    while ($avm =~ /\G\s*(\w+)\s*=>/gc) {
	my $f = $1;
	my $var = "";
	my $val = "";
	my @trail = ();
	$var = $1 if ($avm =~ /\G\s*([A-Z]\w*)\s*(::)?/gc);
	$val = $1 if ($avm =~ /\G\s*([[:alnum:]+-]+)/gc);
	@trail = split(',',$1) if ($avm =~ /\G[\{\[](.*?)[\}\]]/gc);
	$avm =~ /\G\s*,/gc;
	$var="" if ($var && $vars{$var} == 1);
	@trail = ($val) if (!@trail && $val);
	@trail = map( &handle_value($_), @trail);
	my $comp = Forest::Comp->new( 'comp' => $f, 
				      'value' => [@trail] );
	$comp->{'id'} = $var if ($var);
	$fs{$f}=$comp;
    }
    return Forest::Feature->new( 'f' => \%fs );
}
